@@ -9,7 +9,7 @@ from cffi import FFI, VerificationError
 from pywayland.ffi_build import ffi_builder as pywayland_ffi
 from xkbcommon.ffi_build import ffibuilder as xkb_ffi
 
-include_dir = (Path(__file__).parent.parent / "include").resolve()
+include_dir = (Path(__file__).parent / "include").resolve()
 assert include_dir.is_dir(), f"missing {include_dir}"
 
 
@@ -19,6 +19,8 @@ def load_version():
     spec = importlib.util.spec_from_file_location(
         "wlroots.version", os.path.join(dirname, "version.py")
     )
+    assert spec is not None
+    assert spec.loader is not None
     version_module = importlib.util.module_from_spec(spec)
     spec.loader.exec_module(version_module)
     return version_module.version
@@ -477,6 +479,108 @@ struct wlr_gamma_control_manager_v1 *wlr_gamma_control_manager_v1_create(
     struct wl_display *display);
 """
 
+# types/wlr_idle.h
+CDEF += """
+struct wlr_idle {
+    struct wl_global *global;
+    struct wl_list idle_timers; // wlr_idle_timeout::link
+    struct wl_event_loop *event_loop;
+    bool enabled;
+
+    struct wl_listener display_destroy;
+    struct {
+        struct wl_signal activity_notify;
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle_timeout {
+    struct wl_resource *resource;
+    struct wl_list link;
+    struct wlr_seat *seat;
+
+    struct wl_event_source *idle_source;
+    bool idle_state;
+    bool enabled;
+    uint32_t timeout; // milliseconds
+
+    struct {
+        struct wl_signal idle;
+        struct wl_signal resume;
+        struct wl_signal destroy;
+    } events;
+
+    struct wl_listener input_listener;
+    struct wl_listener seat_destroy;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle *wlr_idle_create(struct wl_display *display);
+
+/**
+ * Send notification to restart all timers for the given seat. Called by
+ * compositor when there is an user activity event on that seat.
+ */
+void wlr_idle_notify_activity(struct wlr_idle *idle, struct wlr_seat *seat);
+
+/**
+ * Enable or disable timers for a given idle resource by seat.
+ * Passing a NULL seat means update timers for all seats.
+ */
+void wlr_idle_set_enabled(struct wlr_idle *idle, struct wlr_seat *seat,
+    bool enabled);
+
+/**
+ * Create a new timer on the given seat. The idle event will be called after
+ * the given amount of milliseconds of inactivity, and the resumed event will
+ * be sent at the first user activity after the fired event.
+ */
+struct wlr_idle_timeout *wlr_idle_timeout_create(struct wlr_idle *idle,
+    struct wlr_seat *seat, uint32_t timeout);
+
+void wlr_idle_timeout_destroy(struct wlr_idle_timeout *timeout);
+"""
+
+# types/wlr_input_inhibit_v1.h
+CDEF += """
+struct wlr_idle_inhibit_manager_v1 {
+    struct wl_list inhibitors; // wlr_idle_inhibit_inhibitor_v1::link
+    struct wl_global *global;
+
+    struct wl_listener display_destroy;
+
+    struct {
+        struct wl_signal new_inhibitor;
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle_inhibitor_v1 {
+    struct wlr_surface *surface;
+    struct wl_resource *resource;
+    struct wl_listener surface_destroy;
+
+    struct wl_list link; // wlr_idle_inhibit_manager_v1::inhibitors;
+
+    struct {
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle_inhibit_manager_v1 *wlr_idle_inhibit_v1_create(struct wl_display *display);
+"""
+
 # types/wlr_input_device.h
 CDEF += """
 enum wlr_button_state {
@@ -948,6 +1052,46 @@ void wlr_output_configuration_v1_destroy(
     struct wlr_output_configuration_v1 *config);
 """
 
+# types/wlr_output_powewr_management_v1.h
+CDEF += """
+struct wlr_output_power_manager_v1 {
+    struct wl_global *global;
+    struct wl_list output_powers; // wlr_output_power_v1::link
+
+    struct wl_listener display_destroy;
+
+    struct {
+        struct wl_signal set_mode; // wlr_output_power_v1_set_mode_event
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_output_power_v1 {
+    struct wl_resource *resource;
+    struct wlr_output *output;
+    struct wlr_output_power_manager_v1 *manager;
+    struct wl_list link;
+
+    struct wl_listener output_destroy_listener;
+    struct wl_listener output_commit_listener;
+
+    void *data;
+    ...;
+};
+
+struct wlr_output_power_v1_set_mode_event {
+    struct wlr_output *output;
+    enum zwlr_output_power_v1_mode mode;
+    ...;
+};
+
+struct wlr_output_power_manager_v1 *wlr_output_power_manager_v1_create(
+    struct wl_display *display);
+"""
+
 # types/wlr_pointer.h
 CDEF += """
 struct wlr_pointer {
@@ -2127,6 +2271,8 @@ SOURCE = """
 #include <wlr/types/wlr_data_device.h>
 #include <wlr/types/wlr_foreign_toplevel_management_v1.h>
 #include <wlr/types/wlr_gamma_control_v1.h>
+#include <wlr/types/wlr_idle.h>
+#include <wlr/types/wlr_idle_inhibit_v1.h>
 #include <wlr/types/wlr_input_inhibitor.h>
 #include <wlr/types/wlr_keyboard.h>
 #include <wlr/types/wlr_layer_shell_v1.h>
@@ -2136,6 +2282,7 @@ SOURCE = """
 #include <wlr/types/wlr_output_damage.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_output_management_v1.h>
+#include <wlr/types/wlr_output_power_management_v1.h>
 #include <wlr/types/wlr_pointer_constraints_v1.h>
 #include <wlr/types/wlr_primary_selection_v1.h>
 #include <wlr/types/wlr_relative_pointer_v1.h>
