@@ -1,12 +1,14 @@
 # Copyright (c) 2019 Sean Vig
 
-from typing import Optional, Tuple
+from __future__ import annotations
+
 from weakref import WeakKeyDictionary
 
 from pywayland.server import Display, Signal
 from pywayland.protocol.wayland import WlSeat
 
-from wlroots import ffi, lib, Ptr
+from wlroots import ffi, PtrHasData, lib, Ptr
+from .data_device_manager import Drag
 from .input_device import ButtonState, InputDevice
 from .keyboard import Keyboard, KeyboardModifiers, KeyboardKeyEvent
 from .pointer import AxisSource, AxisOrientation
@@ -16,12 +18,12 @@ _weakkeydict: WeakKeyDictionary = WeakKeyDictionary()
 
 
 class KeyboardGrab(Ptr):
-    def __init__(self, seat: "Seat") -> None:
+    def __init__(self, seat: Seat) -> None:
         """Setup the keyboard grab"""
         self._ptr = ffi.new("struct wlr_seat_keyboard_grab *")
         self._seat = seat
 
-    def __enter__(self) -> "KeyboardGrab":
+    def __enter__(self) -> KeyboardGrab:
         """State the keyboard grab"""
         lib.wlr_seat_keyboard_start_grab(self._seat._ptr, self._ptr)
         return self
@@ -31,7 +33,7 @@ class KeyboardGrab(Ptr):
         lib.wlr_seat_keyboard_end_grab(self._seat)
 
 
-class Seat(Ptr):
+class Seat(PtrHasData):
     def __init__(self, display: Display, name: str) -> None:
         """Allocates a new seat and adds a seat global to the display
 
@@ -57,6 +59,13 @@ class Seat(Ptr):
             ptr=ffi.addressof(self._ptr.events.keyboard_grab_end)
         )
 
+        self.touch_grab_begin_event = Signal(
+            ptr=ffi.addressof(self._ptr.events.touch_grab_begin)
+        )
+        self.touch_grab_end_event = Signal(
+            ptr=ffi.addressof(self._ptr.events.touch_grab_end)
+        )
+
         self.request_set_cursor_event = Signal(
             ptr=ffi.addressof(self._ptr.events.request_set_cursor),
             data_wrapper=PointerRequestSetCursorEvent,
@@ -86,17 +95,20 @@ class Seat(Ptr):
             ptr=ffi.addressof(self._ptr.events.request_start_drag),
             data_wrapper=RequestStartDragEvent,
         )
-        self.start_drag_event = Signal(ptr=ffi.addressof(self._ptr.events.start_drag))
+        self.start_drag_event = Signal(
+            ptr=ffi.addressof(self._ptr.events.start_drag),
+            data_wrapper=Drag,
+        )
 
     @property
-    def pointer_state(self) -> "SeatPointerState":
+    def pointer_state(self) -> SeatPointerState:
         """The pointer state associated with the seat"""
         pointer_state_ptr = ffi.addressof(self._ptr.pointer_state)
         _weakkeydict[pointer_state_ptr] = self._ptr
         return SeatPointerState(pointer_state_ptr)
 
     @property
-    def keyboard_state(self) -> "SeatKeyboardState":
+    def keyboard_state(self) -> SeatKeyboardState:
         """The keyboard state associated with the seat"""
         keyboard_state_ptr = ffi.addressof(self._ptr.keyboard_state)
         _weakkeydict[keyboard_state_ptr] = self._ptr
@@ -141,9 +153,21 @@ class Seat(Ptr):
         return lib.wlr_seat_pointer_surface_has_focus(self._ptr, surface._ptr)
 
     def pointer_clear_focus(self) -> None:
-        """Clear the focused surface for the pointer and leave all entered surfaces"""
+        """Clear the focused surface for the pointer and leave all entered surfaces.
+
+        This function does not respect pointer grabs: you probably want
+        `pointer_notify_clear_focus()` instead.
+        """
+
         return lib.wlr_seat_pointer_clear_focus(self._ptr)
 
+    def pointer_notify_clear_focus(self) -> None:
+        """Notify the seat of a pointer leave event to the currently-focused surface.
+
+        Defers to any grab of the pointer.
+        """
+        return lib.wlr_seat_pointer_notify_clear_focus(self._ptr)
+
     def pointer_notify_enter(
         self, surface: Surface, surface_x: float, surface_y: float
     ) -> None:
@@ -258,6 +282,41 @@ class Seat(Ptr):
         """Whether or not the keyboard has a grab other than the default grab"""
         return lib.wlr_seat_keyboard_has_grab(self._ptr)
 
+    def touch_notify_down(
+        self,
+        surface: Surface,
+        time_msec: int,
+        touch_id: int,
+        surface_x: float,
+        surface_y: float,
+    ) -> None:
+        """
+        Notify the seat of a touch down on the given surface. Defers to any grab of the
+        touch device.
+        """
+        lib.wlr_seat_touch_notify_down(
+            self._ptr, surface._ptr, time_msec, touch_id, surface_x, surface_y
+        )
+
+    def touch_notify_up(self, time_msec: int, touch_id: int) -> None:
+        """
+        Notify the seat that the touch point given by `touch_id` is up. Defers to any
+        grab of the touch device.
+        """
+        lib.wlr_seat_touch_notify_up(self._ptr, time_msec, touch_id)
+
+    def touch_notify_motion(
+        self, time_msec: int, touch_id: int, surface_x: float, surface_y: float
+    ) -> None:
+        """
+        Notify the seat that the touch point given by `touch_id` has moved. Defers to
+        any grab of the touch device. The seat should be notified of touch motion even
+        if the surface is not the owner of the touch point for processing by grabs.
+        """
+        lib.wlr_seat_touch_notify_motion(
+            self._ptr, time_msec, touch_id, surface_x, surface_y
+        )
+
     def set_selection(self, source, serial: int) -> None:
         """Sets the current selection for the seat
 
@@ -271,7 +330,18 @@ class Seat(Ptr):
             # TODO: wrap source in a data source
             lib.wlr_seat_set_selection(self._ptr, source, serial)
 
-    def __enter__(self) -> "Seat":
+    def validate_pointer_grab_serial(self, origin: Surface, serial: int) -> bool:
+        """Check whether this serial is valid to start a pointer grab action."""
+        return lib.wlr_seat_validate_pointer_grab_serial(self._ptr, origin._ptr, serial)
+
+    def start_pointer_drag(self, drag: Drag, serial: int) -> None:
+        """
+        Starts a pointer drag on the seat. This starts implicit keyboard and pointer
+        grabs.
+        """
+        lib.wlr_seat_start_pointer_drag(self._ptr, drag._ptr, serial)
+
+    def __enter__(self) -> Seat:
         """Context manager to clean up the seat"""
         return self
 
@@ -296,7 +366,7 @@ class PointerRequestSetCursorEvent(Ptr):
         return self._ptr.serial
 
     @property
-    def hotspot(self) -> Tuple[int, int]:
+    def hotspot(self) -> tuple[int, int]:
         return self._ptr.hotspot_x, self._ptr.hotspot_y
 
 
@@ -326,7 +396,17 @@ class RequestStartDragEvent(Ptr):
     def __init__(self, ptr) -> None:
         self._ptr = ffi.cast("struct wlr_seat_request_start_drag_event *", ptr)
 
-    # TODO
+    @property
+    def drag(self) -> Drag:
+        return Drag(self._ptr.drag)
+
+    @property
+    def origin(self) -> Surface:
+        return Surface(self._ptr.origin)
+
+    @property
+    def serial(self) -> int:
+        return self._ptr.serial
 
 
 class PointerFocusChangeEvent(Ptr):
@@ -354,7 +434,7 @@ class SeatPointerState(Ptr):
         )
 
     @property
-    def focused_surface(self) -> Optional[Surface]:
+    def focused_surface(self) -> Surface | None:
         """The surface that currently has keyboard focus"""
         focused_surface = self._ptr.focused_surface
         if focused_surface == ffi.NULL:
@@ -373,7 +453,7 @@ class SeatKeyboardState(Ptr):
         )
 
     @property
-    def focused_surface(self) -> Optional[Surface]:
+    def focused_surface(self) -> Surface | None:
         """The surface that is currently focused"""
         focused_surface = self._ptr.focused_surface
         if focused_surface == ffi.NULL:
